VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "UsbFlashSerial"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit
'==============
'UsbFlashSerial
'==============
'
'Per Emmet Gray: "How to get the Serial Number from a USB disk" at:
'
'   http://emmet-gray.com/Articles/USB_SerialNumbers.htm
'
'Converted from VB.Net to VB6 with cleanup.  I hope I didn't add
'bugs by trying to tidy up the code, but I think I've found all of
'his handle leaks and cleaned out redundant logic.  The method
'used appears to return valid information only for USB Flash drives,
'not USB hard drives, so we filter for "removable" drives.
'
'
'PUBLIC METHODS, PROPERTIES, EVENTS, ETC.
'----------------------------------------
'
'There is a single method here:
'
'   Function Lookup(ByVal DriveLetter As String) As String
'
'This returns the drive serial number of the USB drive specified by
'DriveLetter or an empty string on any error (API call errors, not
'a USB drive, bad DriveLetter argument, etc.).
'
'Driveletter must be at least the drive letter itself but only the
'first character of the value is used so even a path may be passed.
'

Private Const FILE_SHARE_NONE As Long = 0
Private Const OPEN_EXISTING As Long = 3
Private Const INVALID_HANDLE_VALUE As Long = -1

Private Const ERROR_INSUFFICIENT_BUFFER As Long = 122

Private Const GUID_DEVINTERFACE_DISK As String = "{53F56307-B6BF-11D0-94F2-00A0C91EFB8B}"
Private Const GUID_DEVINTERFACE_USB_HUB As String = "{F18A0E88-C30C-11D0-8815-00A0C906BED8}"
Private Const DIGCF_PRESENT As Long = &H2
Private Const DIGCF_ALLCLASSES As Long = &H4
Private Const DIGCF_DEVICEINTERFACE As Long = &H10
Private Const SPDRP_DRIVER As Long = &H9
Private Const USB_DEVICE_DESCRIPTOR_TYPE As Integer = &H1
Private Const USB_STRING_DESCRIPTOR_TYPE As Integer = &H3

Private Const PNP_ENUMID_SYMBOLIC_USB As String = "USB"

Private Const IOCTL_STORAGE_GET_DEVICE_NUMBER As Long = &H2D1080
Private Const IOCTL_USB_GET_DESCRIPTOR_FROM_NODE_CONNECTION As Long = &H220410
Private Const IOCTL_USB_GET_NODE_INFORMATION As Long = &H220408
Private Const IOCTL_USB_GET_NODE_CONNECTION_INFORMATION_EX = &H220448
Private Const IOCTL_USB_GET_NODE_CONNECTION_DRIVERKEY_NAME As Long = &H220420

Private Type GUID
    Bytes(15) As Byte
End Type

Private Type SETUP_PACKET
    bmRequest As Byte
    bRequest As Byte
    wValue As Integer
    wIndex As Integer
    wLength As Integer
End Type

Private Type SP_DEVICE_INTERFACE_DATA
    cbSize As Long
    InterfaceClassGuid As GUID
    Flags As Long
    Reserved As Long
End Type

Private Type SP_DEVICE_INTERFACE_DETAIL_DATA
    cbSize As Long
    DevicePath(511) As Byte
End Type

Private Type SP_DEVINFO_DATA
    cbSize As Long
    ClassGuid As GUID
    DevInst As Long
    Reserved As Long
End Type

Private Type STORAGE_DEVICE_NUMBER
    DeviceType As Long
    DeviceNumber As Long
    PartitionNumber As Long
End Type

Private Type USB_DEVICE_DESCRIPTOR
    bLength As Byte
    bDescriptorType As Byte
    bcdUSB As Integer
    bDeviceClass As Byte
    bDeviceSubClass As Byte
    bDeviceProtocol As Byte
    bMaxPacketSize0 As Byte
    idVendor As Integer
    idProduct As Integer
    bcdDevice As Integer
    iManufacturer As Byte
    iProduct As Byte
    iSerialNumber As Byte
    bNumConfigurations As Byte
End Type

Private Type USB_DESCRIPTOR_REQUEST_4DEVICE
    ConnectionIndex As Long
    SetupPacket As SETUP_PACKET
    Data As USB_DEVICE_DESCRIPTOR
End Type

Private Type USB_STRING_DESCRIPTOR
    bLength As Byte
    bDescriptorType As Byte
    bString(63) As Byte
End Type

Private Type USB_DESCRIPTOR_REQUEST_4STRING
    ConnectionIndex As Long
    SetupPacket As SETUP_PACKET
    Data As USB_STRING_DESCRIPTOR
End Type

Private Enum USB_HUB_NODE
    UsbHub
    UsbMIParent
End Enum

Private Type USB_HUB_DESCRIPTOR
    bDescriptorLength As Byte
    bDescriptorType As Byte
    bNumberOfPorts As Byte
    'Here we have to fiddle because we need a packed structure:
    wHubCharacteristics(1) As Byte 'As Integer
    bPowerOnToPowerGood As Byte
    bHubControlCurrent As Byte
    bRemoveAndPowerMask(63) As Byte
End Type

Private Type USB_HUB_INFORMATION
    HubDescriptor As USB_HUB_DESCRIPTOR
    HubIsBusPowered As Byte
End Type

Private Type USB_NODE_INFORMATION
    NodeType As USB_HUB_NODE
    HubInformation As USB_HUB_INFORMATION
End Type

Private Enum USB_CONNECTION_STATUS
    NoDeviceConnected
    DeviceConnected
    DeviceFailedEnumeration
    DeviceGeneralFailure
    DeviceCausedOvercurrent
    DeviceNotEnoughPower
    DeviceNotEnoughBandwidth
    DeviceHubNestedTooDeeply
    DeviceInLegacyHub
End Enum

Private Type USB_NODE_CONNECTION_INFORMATION_EX
    ConnectionIndex As Long
    DeviceDescriptor As USB_DEVICE_DESCRIPTOR
    CurrentConfigurationValue As Byte
    Speed As Byte
    DeviceIsHub As Byte
    'Here we have to fiddle because we need a packed structure:
    DeviceAddress(1) As Byte 'As Integer
    NumberOfOpenPipes(3) As Byte 'As Long
    ConnectionStatus(3) As Byte 'As USB_CONNECTION_STATUS (enum, but byte 0 is good enough).
    'USB_PIPE_INFO PipeList[0] (0 length so we'll comment this out).
End Type

Private Type USB_NODE_CONNECTION_DRIVERKEY_NAME
    ConnectionIndex As Long
    ActualLength As Long
    DriverKeyName(511) As Byte
End Type

Private Enum REG_TYPES
    REG_BINARY = 3
    REG_DWORD = 4
    REG_DWORD_BIG_ENDIAN = 5
    REG_EXPAND_SZ = 2
    REG_MULTI_SZ = 7
    REG_SZ = 1
End Enum

Private Enum DriveTypeEnum
    dtUnknown = 0
    dtNoRootDir
    dtRemovable
    dtFixed
    dtRemote
    dtCDROM
    dtRAMDisk
    [_dtLAST] = dtRAMDisk
End Enum

Private Declare Function CloseHandle Lib "kernel32" (ByVal hObject As Long) As Long

Private Declare Function CM_Get_Device_ID Lib "cfgmgr32" _
    Alias "CM_Get_Device_IDA" ( _
    ByVal dnDevInst As Long, _
    ByRef Buffer As Byte, _
    ByVal BufferLen As Long, _
    ByVal ulFlags As Long) As Long

Private Declare Function CM_Get_Device_ID_Size Lib "cfgmgr32" ( _
    ByRef pulLen As Long, _
    ByVal dnDevInst As Long, _
    ByVal ulFlags As Long) As Long

Private Declare Function CM_Get_Parent Lib "cfgmgr32" ( _
    ByRef pdnDevInst As Long, _
    ByVal dnDevInst As Long, _
    ByVal ulFlags As Long) As Long

Private Declare Function CreateFile Lib "kernel32" Alias "CreateFileA" ( _
    ByVal lpFileName As String, _
    ByVal dwDesiredAccess As Long, _
    ByVal dwShareMode As Long, _
    ByVal lpSecurityAttributes As Long, _
    ByVal dwCreationDisposition As Long, _
    ByVal dwFlagsAndAttributes As Long, _
    ByVal hTemplateFile As Long) As Long

Private Declare Function DeviceIoControl Lib "kernel32" ( _
    ByVal hDevice As Long, _
    ByVal dwIoControlCode As Long, _
    ByVal lpInBuffer As Long, _
    ByVal nInBufferSize As Long, _
    ByVal lpOutBuffer As Long, _
    ByVal nOutBufferSize As Long, _
    ByRef lpBytesReturned As Long, _
    ByVal lpOverlapped As Long) As Long

Private Declare Function GetDriveType Lib "kernel32" _
    Alias "GetDriveTypeA" ( _
    ByVal RootPathName As String) As DriveTypeEnum

Private Declare Function GUIDFromString Lib "shlwapi" Alias "#269" ( _
    ByVal psz As String, _
    ByRef pguid As GUID) As Long

Private Declare Function SetupDiDestroyDeviceInfoList Lib "setupapi" ( _
    ByVal DeviceInfoSet As Long) As Long

Private Declare Function SetupDiEnumDeviceInfo Lib "setupapi" ( _
    ByVal DeviceInfoSet As Long, _
    ByVal MemberIndex As Long, _
    ByRef DeviceInfoData As SP_DEVINFO_DATA) As Long

Private Declare Function SetupDiEnumDeviceInterfaces Lib "setupapi" ( _
    ByVal DeviceInfoSet As Long, _
    ByVal DeviceInfoData As Long, _
    ByRef InterfaceClassGuid As GUID, _
    ByVal MemberIndex As Long, _
    ByRef DeviceInterfaceData As SP_DEVICE_INTERFACE_DATA) As Long

Private Declare Function SetupDiGetClassDevs Lib "setupapi" Alias "SetupDiGetClassDevsA" ( _
    ByVal pClassGuid As Long, _
    ByVal Enumerator As String, _
    ByVal hWndParent As Long, _
    ByVal Flags As Long) As Long

Private Declare Function SetupDiGetDeviceInstanceId Lib "setupapi" _
    Alias "SetupDiGetDeviceInstanceIdA" ( _
    ByVal DeviceInfoSet As Long, _
    ByRef DeviceInfoData As SP_DEVINFO_DATA, _
    ByRef DeviceInstanceId As Byte, _
    ByVal DeviceInstanceIdSize As Long, _
    ByRef RequiredSize As Long) As Long

Private Declare Function SetupDiGetDeviceInterfaceDetail Lib "setupapi" _
    Alias "SetupDiGetDeviceInterfaceDetailA" ( _
    ByVal DeviceInfoSet As Long, _
    ByRef DeviceInterfaceData As SP_DEVICE_INTERFACE_DATA, _
    ByRef DeviceInterfaceDetailData As SP_DEVICE_INTERFACE_DETAIL_DATA, _
    ByVal DeviceInterfaceDetailDataSize As Long, _
    ByRef RequiredSize As Long, _
    ByRef DeviceInfoData As SP_DEVINFO_DATA) As Long

Private Declare Function SetupDiGetDeviceRegistryProperty Lib "setupapi" _
    Alias "SetupDiGetDeviceRegistryPropertyA" ( _
    ByVal DeviceInfoSet As Long, _
    ByRef DeviceInfoData As SP_DEVINFO_DATA, _
    ByVal Property As Long, _
    ByRef PropertyRegDataType As REG_TYPES, _
    ByRef PropertyBuffer As Byte, _
    ByVal PropertyBufferSize As Long, _
    ByRef RequiredSize As Long) As Long

Public Function Lookup(ByVal DriveLetter As String) As String
    Dim HubDevicePath As String
    Dim InstanceID As String
    Dim hHub As Long
    Dim UsbNodeInfo As USB_NODE_INFORMATION
    Dim BytesReturned As Long
    Dim I As Byte
    Dim DriverKeyName As String
    Dim InstanceIDByKeyName As String
    
    DriveLetter = Left$(DriveLetter, 1) & ":"
    If GetDriveType(DriveLetter & "\") = dtRemovable Then
        If FindDiskDevice(DriveLetter, HubDevicePath, InstanceID) Then
            If UCase$(Left$(InstanceID, 4)) = "USB\" Then
                hHub = CreateFile(HubDevicePath, _
                                  0, _
                                  FILE_SHARE_NONE, _
                                  0, _
                                  OPEN_EXISTING, _
                                  0, _
                                  0)
                If hHub <> INVALID_HANDLE_VALUE Then
                    UsbNodeInfo.NodeType = UsbHub
                    If DeviceIoControl(hHub, _
                                       IOCTL_USB_GET_NODE_INFORMATION, _
                                       VarPtr(UsbNodeInfo), _
                                       Len(UsbNodeInfo), _
                                       VarPtr(UsbNodeInfo), _
                                       Len(UsbNodeInfo), _
                                       BytesReturned, _
                                       0) Then
                        For I = 1 To UsbNodeInfo.HubInformation.HubDescriptor.bNumberOfPorts
                            DriverKeyName = GetDriverKeyName(hHub, I)
                            If Len(DriverKeyName) > 0 Then
                                InstanceIDByKeyName = FindInstanceIDByKeyName(DriverKeyName)
                                If Len(InstanceIDByKeyName) > 0 Then
                                    If InstanceIDByKeyName = InstanceID Then
                                        Lookup = GetSerialNumberString(hHub, I)
                                        Exit For
                                    End If
                                End If
                            End If
                        Next
                    End If
                    CloseHandle hHub
                End If
            End If
        End If
    End If
End Function

Private Function FindDiskDevice( _
    ByVal DriveLetter As String, _
    ByRef HubDevicePath As String, _
    ByRef InstanceID As String) As Boolean
    
    Dim DeviceNumber As Long
    Dim DiskGUID As GUID
    Dim hDevInfo As Long
    Dim Success As Boolean
    Dim I As Long
    Dim DevIntfData As SP_DEVICE_INTERFACE_DATA
    Dim DevInfoData As SP_DEVINFO_DATA
    Dim DevIntfDetailData As SP_DEVICE_INTERFACE_DETAIL_DATA
    Dim RequiredSize As Long
    Dim DevicePath As String
    Dim diParent As Long
    Dim Size As Long
    Dim Bytes() As Byte
    Dim HubID As String
    
    DeviceNumber = GetDeviceNumber("\\.\" & DriveLetter)
    If DeviceNumber >= 0 Then
        GUIDFromString GUID_DEVINTERFACE_DISK, DiskGUID
        hDevInfo = SetupDiGetClassDevs(VarPtr(DiskGUID), _
                                       vbNullString, _
                                       0, _
                                       DIGCF_PRESENT Or DIGCF_DEVICEINTERFACE)
        If hDevInfo <> INVALID_HANDLE_VALUE Then
            Do
                DevIntfData.cbSize = Len(DevIntfData)
                Success = SetupDiEnumDeviceInterfaces(hDevInfo, _
                                                      0, _
                                                      DiskGUID, _
                                                      I, _
                                                      DevIntfData)
                If Success Then
                    DevInfoData.cbSize = Len(DevInfoData)
                    'Oddity here: We must set the size to everything prior to
                    '.DevicePath, plus 1 byte for the (vestigial, nominal?)
                    '.DevicePath!
                    DevIntfDetailData.cbSize = Len(DevIntfDetailData.cbSize) + 1
                    If SetupDiGetDeviceInterfaceDetail(hDevInfo, _
                                                       DevIntfData, _
                                                       DevIntfDetailData, _
                                                       Len(DevIntfDetailData), _
                                                       RequiredSize, _
                                                       DevInfoData) Then
                        'RequiredSize is the entire structure, which includes the
                        'other structure members (4 bytes) and a terminating NUL (1).
                        DevicePath = _
                            StrConv(LeftB$(DevIntfDetailData.DevicePath, RequiredSize - 4 - 1), _
                                    vbUnicode)
'                        '@@@ Appears completely unnecessary, but left here
'                        '@@@ in case someone finds it that it is needed.
'                        If Right$(DevicePath, 1) = "\" Then
'                            DevicePath = Left$(DevicePath, Len(DevicePath) - 1)
'                        End If
                        If GetDeviceNumber(DevicePath) = DeviceNumber Then
                            'This is the device we're looking for.  Back up through
                            'the hierarchy to obtain the Instance ID and the
                            'Hub's ID.
                            CM_Get_Parent diParent, DevInfoData.DevInst, 0
                            CM_Get_Device_ID_Size Size, diParent, 0
                            ReDim Bytes(Size) 'Note: 1 extra for terminating NUL not included in Size.
                            CM_Get_Device_ID diParent, Bytes(0), Size + 1, 0
                            InstanceID = StrConv(LeftB$(Bytes, Size), vbUnicode)
                            CM_Get_Parent diParent, diParent, 0
                            CM_Get_Device_ID_Size Size, diParent, 0
                            ReDim Bytes(Size)
                            CM_Get_Device_ID diParent, Bytes(0), Size + 1, 0
                            HubID = StrConv(LeftB$(Bytes, Size), vbUnicode)
                            Erase Bytes
                            HubDevicePath = "\\.\" & Replace$(HubID, "\", "#") _
                                          & "#" & GUID_DEVINTERFACE_USB_HUB
                            FindDiskDevice = True
                            Exit Do
                        End If
                    Else
                        Exit Do
                    End If
                End If
                I = I + 1
            Loop While Success
        
            SetupDiDestroyDeviceInfoList hDevInfo
        End If
    End If
End Function

Private Function GetDeviceNumber(ByVal DevicePath As String) As Long
    Dim hDevice As Long
    Dim StorDevNum As STORAGE_DEVICE_NUMBER
    Dim BytesReturned As Long

    hDevice = CreateFile(DevicePath, _
                         0, _
                         FILE_SHARE_NONE, _
                         0, _
                         OPEN_EXISTING, _
                         0, _
                         0)
    If hDevice <> INVALID_HANDLE_VALUE Then
        If DeviceIoControl(hDevice, _
                           IOCTL_STORAGE_GET_DEVICE_NUMBER, _
                           0, _
                           0, _
                           VarPtr(StorDevNum), _
                           Len(StorDevNum), _
                           BytesReturned, _
                           0) Then
            GetDeviceNumber = StorDevNum.DeviceType * &H10000 + StorDevNum.DeviceNumber
        Else
            GetDeviceNumber = -1
        End If
        CloseHandle hDevice
    Else
        GetDeviceNumber = -1
    End If
End Function

Private Function GetDriverKeyName(ByVal hHub As Long, ByVal PortNumber As Byte) As String
    Dim Result As Long
    Dim UsbConnInfoEx As USB_NODE_CONNECTION_INFORMATION_EX
    Dim UsbConnDriverkeyName As USB_NODE_CONNECTION_DRIVERKEY_NAME
    Dim BytesReturned As Long

    UsbConnInfoEx.ConnectionIndex = PortNumber
    Result = DeviceIoControl(hHub, _
                             IOCTL_USB_GET_NODE_CONNECTION_INFORMATION_EX, _
                             VarPtr(UsbConnInfoEx), _
                             Len(UsbConnInfoEx), _
                             VarPtr(UsbConnInfoEx), _
                             Len(UsbConnInfoEx), _
                             BytesReturned, _
                             0)
    If Result Then
        If UsbConnInfoEx.ConnectionStatus(0) = DeviceConnected Then
            UsbConnDriverkeyName.ConnectionIndex = PortNumber
            Result = DeviceIoControl(hHub, _
                                     IOCTL_USB_GET_NODE_CONNECTION_DRIVERKEY_NAME, _
                                     VarPtr(UsbConnDriverkeyName), _
                                     Len(UsbConnDriverkeyName), _
                                     VarPtr(UsbConnDriverkeyName), _
                                     Len(UsbConnDriverkeyName), _
                                     BytesReturned, _
                                     0)
            If Result Then
                'This is an oddball:  Even though we're making an ANSI call
                'we'll always get Unicode back here.
                '
                'We can't use either BytesReturned or .ActualLength to dig
                'out the .DriverKeyName.  They will often (always?) be too
                'large even allowing for the previous members of the structure.
                'It almost seems as if .DriverKeyName is terminated by *two*
                'Unicode NULs.  Instead we'll scan for a NUL.
                GetDriverKeyName = Left$(UsbConnDriverkeyName.DriverKeyName, _
                                         InStr(UsbConnDriverkeyName.DriverKeyName, _
                                               vbNullChar) - 1)
            End If
        End If
    End If
End Function

Private Function FindInstanceIDByKeyName(ByVal DriverKeyName As String) As String
    Dim hDevInfo As Long
    Dim Success As Boolean
    Dim I As Long
    Dim KeyName As String
    Dim DevInfoData As SP_DEVINFO_DATA
    Dim RegType As REG_TYPES
    Dim Bytes() As Byte
    Dim RequiredSize As Long
    
    hDevInfo = SetupDiGetClassDevs(0, _
                                   PNP_ENUMID_SYMBOLIC_USB, _
                                   0, _
                                   DIGCF_PRESENT Or DIGCF_ALLCLASSES)
    If hDevInfo <> INVALID_HANDLE_VALUE Then
        Do
            DevInfoData.cbSize = Len(DevInfoData)
            Success = SetupDiEnumDeviceInfo(hDevInfo, I, DevInfoData)
            If Success Then
                KeyName = ""
                SetupDiGetDeviceRegistryProperty hDevInfo, _
                                                 DevInfoData, _
                                                 SPDRP_DRIVER, _
                                                 RegType, _
                                                 ByVal 0&, _
                                                 0, _
                                                 RequiredSize
                If Err.LastDllError = ERROR_INSUFFICIENT_BUFFER Then
                    ReDim Bytes(RequiredSize - 1)
                    If SetupDiGetDeviceRegistryProperty(hDevInfo, _
                                                        DevInfoData, _
                                                        SPDRP_DRIVER, _
                                                        RegType, _
                                                        Bytes(0), _
                                                        RequiredSize, _
                                                        ByVal 0&) Then
                        KeyName = StrConv(LeftB$(Bytes, RequiredSize - 1), vbUnicode)
                    End If
                    Erase Bytes
                Else
                    Exit Do
                End If
                If KeyName = DriverKeyName Then
                    SetupDiGetDeviceInstanceId hDevInfo, _
                                               DevInfoData, _
                                               ByVal 0&, _
                                               0, _
                                               RequiredSize
                    If Err.LastDllError = ERROR_INSUFFICIENT_BUFFER Then
                        ReDim Bytes(RequiredSize - 1)
                        If SetupDiGetDeviceInstanceId(hDevInfo, _
                                                      DevInfoData, _
                                                      Bytes(0), _
                                                      RequiredSize, _
                                                      ByVal 0&) Then
                            FindInstanceIDByKeyName = _
                                StrConv(LeftB$(Bytes, RequiredSize - 1), vbUnicode)
                        End If
                    Else
                        Exit Do
                    End If
                    Exit Do
                End If
            End If
            I = I + 1
        Loop While Success
        
        SetupDiDestroyDeviceInfoList hDevInfo
    End If
End Function

Private Function GetSerialNumberString(ByVal hHub As Long, ByVal PortNumber As Byte) As String
    Dim UsbDescReq4Device As USB_DESCRIPTOR_REQUEST_4DEVICE
    Dim UsbDescReq4String As USB_DESCRIPTOR_REQUEST_4STRING
    Dim BytesReturned
    
    With UsbDescReq4Device
        .ConnectionIndex = PortNumber
        .SetupPacket.wValue = USB_DEVICE_DESCRIPTOR_TYPE * &H100 '+ 0 here, i.e. Index = 0 per USB Spec.
        .SetupPacket.wLength = Len(.Data)
    End With
    If DeviceIoControl(hHub, _
                       IOCTL_USB_GET_DESCRIPTOR_FROM_NODE_CONNECTION, _
                       VarPtr(UsbDescReq4Device), _
                       Len(UsbDescReq4Device), _
                       VarPtr(UsbDescReq4Device), _
                       Len(UsbDescReq4Device), _
                       BytesReturned, _
                       0) Then
        With UsbDescReq4String
            .ConnectionIndex = PortNumber
            .SetupPacket.wValue = USB_STRING_DESCRIPTOR_TYPE * &H100 _
                                + UsbDescReq4Device.Data.iSerialNumber
            .SetupPacket.wLength = Len(.Data)
            'Here &H409 = LCID for US English, 0 = current system language.  We'll
            'comment this out since it was zeroed when we created it here:
            '.SetupPacket.wIndex = 0
        End With
        If DeviceIoControl(hHub, _
                           IOCTL_USB_GET_DESCRIPTOR_FROM_NODE_CONNECTION, _
                           VarPtr(UsbDescReq4String), _
                           Len(UsbDescReq4String), _
                           VarPtr(UsbDescReq4String), _
                           Len(UsbDescReq4String), _
                           BytesReturned, _
                           0) Then
            GetSerialNumberString = Left$(UsbDescReq4String.Data.bString, _
                                          UsbDescReq4String.Data.bLength \ 2 - 1)
        End If
    End If
End Function


